#!/usr/bin/env python

import sys

# Method 1: Test whether s.decode() returns surrogate pair or single char
s = b'\\U00010142'
c = s.decode('unicode-escape')
if len(c) == 1:
    size1 = 'wide'
elif len(c) == 2:
    size1 = 'narrow'
else:
    raise AssertionError(
        f"python_unicodesize.py: Length of Unicode character {s} returned from "
        f"s.decode() must be 1 or 2 but is: {len(c)}")

# Method 2: Test whether unichr() fails
try:
    if sys.version_info[0] == 2:
        unichr(65858)  # U+10142
    else:
        chr(65858)  # U+10142
except ValueError:
    size2 = 'narrow'
else:
    size2 = 'wide'

if size1 != size2:
    raise AssertionError(
        "python_unicodesize.py: Different results for unicode size: "
        f"s.decode() method: {size1}, unichr() method: {size2}")

print(f"Unicode size of Python environment: {size1}")



#!/usr/bin/env python
#
# Build the LEX/YACC table modules for the MOF compiler, in the PyWBEM
# installation directory. The installation directory is used by removing
# the directory of this script from the Python module search path.

import sys
import os
from pprint import pprint

myname = os.path.basename(sys.argv[0])

DEBUG = False   # Enable to get debug info
VERBOSE = True  # TODO: Pass as command line option


def env_info():
    print(f"{myname}: Platform: {sys.platform}")
    print(f"{myname}: Python executable: {sys.executable}")
    print(f"{myname}: Python version: {sys.version}")
    print(f"{myname}: Python module search path:")
    pprint(sys.path)


def main():

    if VERBOSE:
        print(f"{myname}: Rebuilding the pywbem LEX/YACC table modules, "
              "if needed")

    # We want to build the table modules in the target install directory.
    #
    # The _mof_compiler._build() function builds them in the directory where
    # the _mof_compiler module was imported from.
    #
    # This script runs in the directory where the distribution archive has
    # been unpacked to (or during development in the git repo work directory),
    # where the pywbem package is directly available as a subdirectory.
    #
    # Therefore, we need to remove the directory of this script from the Python
    # module search path before importing the _mof_compiler module, in order to
    # import it from the target install directory.
    if VERBOSE:
        print(f"{myname}: Ensuring to build in pywbem install directory")
    my_abs_path = os.path.dirname(os.path.abspath(sys.argv[0]))
    removals = []
    for i, p in enumerate(sys.path):
        abs_p = os.path.abspath(p)
        if abs_p == my_abs_path:
            if DEBUG:
                print(f"{myname}: Debug: Removing from module search path: "
                      f"{abs_p}")
            removals.append(i)
    for i in reversed(removals):
        del sys.path[i]
    if DEBUG:
        print(f"{myname}: Debug: Resulting module search path:")
        pprint(sys.path)

    try:
        import pywbem
        if DEBUG:
            print(f"{myname}: Debug: Pywbem package successfully imported "
                  f"from: {pywbem.__file__}")
    except ImportError as exc:
        print(f"{myname}: Error: Import of pywbem package failed: {exc}")
        env_info()
        return 1

    from pywbem import _mof_compiler
    try:
        _mof_compiler._build(verbose=True)
    except Exception as exc:
        print(f"{myname}: Error: Rebuilding the pywbem LEX/YACC table modules "
              "failed: {exc}")
        env_info()
        return 1

    if VERBOSE:
        print(f"{myname}: Successfully rebuilt the pywbem LEX/YACC table "
              "modules, if needed")
    return 0

if __name__ == '__main__':
    sys.exit(main())



"""
Resource consumption tests for mocked EnumerateInstances and
EnumerateInstanceNames operations.
"""

import pytest

from .random_objects import tst_random_conn
from .memory_utils import total_sizeof
from .resource_measurement import ResourceMeasurement


@pytest.mark.parametrize(
    "num_props, num_insts", [
        (1, 1),
        (10, 100),
    ]
)
def test_enuminst_mock(num_props, num_insts):
    """
    Resource consumption test for EnumerateInstances operation against
    a mock environment.
    """

    conn, cls, _ = tst_random_conn(num_insts, num_props)

    print(f"\nResource consumption test: EnumerateInstances with "
          f"num_props={num_props}, num_insts={num_insts}:")

    with ResourceMeasurement() as rm:
        result = conn.EnumerateInstances(cls.classname)

    result_size = total_sizeof(result)

    print(f"  Elapsed time: {rm.elapsed_time_ms:>12.3f} ms")
    print(f"  CPU time:     {rm.cpu_time_ms:>12.3f} ms")
    print(f"  Peak usage:   {rm.peak_size:>12} B")
    print(f"  Result size:  {result_size:>12} B")


@pytest.mark.parametrize(
    "num_props, num_insts", [
        (1, 1),
        (10, 100),
    ]
)
def test_enuminstnames_mock(num_props, num_insts):
    """
    Resource consumption test for EnumerateInstanceNames operation against
    a mock environment.
    """

    conn, cls, _ = tst_random_conn(num_insts, num_props)

    print(f"\nResource consumption test: EnumerateInstanceNames with "
          f"num_props={num_props}, num_insts={num_insts}:")

    with ResourceMeasurement() as rm:
        result = conn.EnumerateInstanceNames(cls.classname)

    result_size = total_sizeof(result)

    print(f"  Elapsed time: {rm.elapsed_time_ms:>12.3f} ms")
    print(f"  CPU time:     {rm.cpu_time_ms:>12.3f} ms")
    print(f"  Peak usage:   {rm.peak_size:>12} B")
    print(f"  Result size:  {result_size:>12} B")



"""
Test memory leaks for dictionary classes used by pywbem.
"""

from collections import OrderedDict
try:
    from django.utils.datastructures import SortedDict as Django_SortedDict
except ImportError:
    Django_SortedDict = None
import pytest
import yagot

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ..utils import import_installed
pywbem = import_installed('pywbem')
from pywbem._cim_obj import NocaseDict  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


@yagot.garbage_checked()
def test_leaks_OrderedDict_empty():
    """
    Test function with empty OrderedDict object.
    """
    _ = OrderedDict()


@pytest.mark.skipif(
    Django_SortedDict is None,
    reason="django with SortedDict is not installed")
@yagot.garbage_checked()
def test_leaks_Django_SortedDict_empty():
    """
    Test function with empty django SortedDict object.

    Note: Django's SortedDict has been removed from the django package as of
    version 1.9. This test requires django<1.9 to be installed and is skipped
    otherwise.
    """
    _ = Django_SortedDict()


@yagot.garbage_checked()
def test_leaks_NocaseDict_empty():
    """
    Test function with empty NocaseDict object.
    """
    _ = NocaseDict()


@yagot.garbage_checked()
def test_leaks_NocaseDict_one():
    """
    Test function with NocaseDict object containing one string item.
    """
    _ = NocaseDict(a='b')



#!/usr/bin/env python

"""
Test _listener.py module.
"""

import sys
import re
import logging
import threading
from time import time, sleep
from random import randint
try:
    from types import NoneType
except ImportError:
    # On Python <= 3.9
    NoneType = type(None)
import requests
import pytest

from ...utils import post_bsl
from ...elapsed_timer import ElapsedTimer
from ..utils.pytest_extensions import simplified_test_function, \
    log_entry_exit, get_logger

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import WBEMListener, ListenerPortError, \
    ListenerQueueFullError  # noqa: E402
from pywbem._listener import ExceptionHandlingThread, StoppableThread, \
    ServerThread, CallbackThread  # noqa: E402
from pywbem._utils import _format  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name

# Literal form {"blah: 0} faster than dict(blah=0) but same functionality
# pylint: disable=use-dict-literal

# test variables to allow selectively executing tests.
OK = True
RUN = True
FAIL = False

LOGGER = get_logger(__name__)

TESTCASES_WBEMLISTENER_INIT = [

    # Testcases for WBEMListener.__init__()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * init_args: Tuple of positional arguments to WBEMListener().
    #   * init_kwargs: Dict of keyword arguments to WBEMListener().
    #   * exp_attrs: Dict of expected attributes of resulting object.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Success cases
    (
        "Verify order of positional arguments",
        dict(
            init_args=[
                'woot.com',
                6997,
                6998,
                'certfile.pem',
                'keyfile.pem',
            ],
            init_kwargs={},
            exp_attrs=dict(
                host='woot.com',
                http_port=6997,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
        ),
        None, None, True
    ),
    (
        "Verify names of keyword arguments",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=6997,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
                max_ind_queue_size=9,
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=6997,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
                max_ind_queue_size=9

            ),
        ),
        None, None, True
    ),
    (
        "Verify minimal arguments for success with http_port as integer",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=6997,
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=6997,
                https_port=None,
                certfile=None,
                keyfile=None,
            ),
        ),
        None, None, True
    ),
    (
        "Verify minimal arguments for success with http_port as string",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port='6997',
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=6997,
                https_port=None,
                certfile=None,
                keyfile=None,
            ),
        ),
        None, None, True
    ),
    (
        "Verify full arguments for success with http_port",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=6997,
                https_port=None,
                certfile=None,
                keyfile=None,
                max_ind_queue_size=0,
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=6997,
                https_port=None,
                certfile=None,
                keyfile=None,
                max_ind_queue_size=0,

            ),
        ),
        None, None, True
    ),
    (
        "Verify minimal arguments for success with https_port as integer",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=None,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
        ),
        None, None, True
    ),
    (
        "Verify minimal arguments for success with https_port as string",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                https_port='6998',
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=None,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
        ),
        None, None, True
    ),
    (
        "Verify full arguments for success with https_port",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=None,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=None,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
        ),
        None, None, True
    ),
    (
        "Verify use of direct_call argument",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=None,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=None,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
        ),
        None, None, True
    ),
    (
        "Verify valid max_ind_queue_size argument",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=None,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
                max_ind_queue_size=1000,
            ),
            exp_attrs=dict(
                host='woot.com',
                http_port=None,
                https_port=6998,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
                max_ind_queue_size=1000,

            ),
        ),
        None, None, True
    ),

    # Failure cases
    (
        "Verify failure when providing invalid type for http_port",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=12.1,
            ),
            exp_attrs=None,
        ),
        TypeError, None, True
    ),
    (
        "Verify failure when providing invalid type for https_port",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                https_port=12.1,
                certfile='certfile.pem',
                keyfile='keyfile.pem',
            ),
            exp_attrs=None,
        ),
        TypeError, None, True
    ),
    (
        "Verify failure when providing no certfile with https_port",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                https_port=6998,
                certfile=None,
                keyfile='keyfile.pem',
            ),
            exp_attrs=None,
        ),
        ValueError, None, True
    ),
    (
        "Verify failure when providing no keyfile with https_port",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                https_port=6998,
                certfile='certfile.pem',
                keyfile=None,
            ),
            exp_attrs=None,
        ),
        ValueError, None, True
    ),
    (
        "Verify failure when providing no port",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=None,
                https_port=None,
            ),
            exp_attrs=None,
        ),
        ValueError, None, True
    ),
    (
        "Verify failure when providing invalid type max_ind_queue_size",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=None,
                https_port=None,
                max_ind_queue_size="fred",
            ),
            exp_attrs=None,
        ),
        ValueError, None, True
    ),
    (
        "Verify failure when providing invalid max_ind_queue_size integer",
        dict(
            init_args=[],
            init_kwargs=dict(
                host='woot.com',
                http_port=None,
                https_port=None,
                max_ind_queue_size=-1,
            ),
            exp_attrs=None,
        ),
        ValueError, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_WBEMLISTENER_INIT)
@simplified_test_function
@log_entry_exit
def test_WBEMListener_init(testcase, init_args, init_kwargs, exp_attrs):
    """
    Test function for WBEMListener.__init__()
    """

    # The code to be tested
    obj = WBEMListener(*init_args, **init_kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # Verify specified expected attributes
    for attr_name in exp_attrs:
        exp_attr = exp_attrs[attr_name]
        act_attr = getattr(obj, attr_name)
        assert act_attr == exp_attr
        assert isinstance(act_attr, type(exp_attr))

    # Verify attributes not set via init arguments

    assert isinstance(obj.logger, logging.Logger)
    assert re.match(r'pywbem\.listener\.', obj.logger.name)

    assert obj.http_started is False
    assert obj.https_started is False


TESTCASES_WBEMLISTENER_STR = [

    # Testcases for WBEMListener.__str__() / str()

    # Each list item is a testcase tuple with these items:
    # * obj: WBEMListener object to be tested.

    (
        WBEMListener(
            host='woot.com',
            http_port=6997,
            https_port=6998,
            certfile='certfile.pem',
            keyfile='keyfile.pem')
    ),
]


@pytest.mark.parametrize(
    "obj",
    TESTCASES_WBEMLISTENER_STR)
@log_entry_exit
def test_WBEMListener_str(obj):
    """
    Test function for WBEMListener.__str__() / str()
    """

    # The code to be tested
    result = str(obj)

    assert re.match(r'^WBEMListener\(', result)

    exp_host_str = _format('_host={0!A}', obj.host)
    assert exp_host_str in result

    exp_http_port_str = _format('_http_port={0!A}', obj.http_port)
    assert exp_http_port_str in result

    exp_https_port_str = _format('_https_port={0!A}', obj.https_port)
    assert exp_https_port_str in result


TESTCASES_WBEMLISTENER_REPR = [

    # Testcases for WBEMListener.__repr__() / repr()

    # Each list item is a testcase tuple with these items:
    # * obj: WBEMListener object to be tested.

    (
        WBEMListener(
            host='woot.com',
            http_port=6997,
            https_port=6998,
            certfile='certfile.pem',
            keyfile='keyfile.pem')
    ),
]


@pytest.mark.parametrize(
    "obj",
    TESTCASES_WBEMLISTENER_REPR)
@log_entry_exit
def test_WBEMListener_repr(obj):
    """
    Test function for WBEMListener.__repr__() / repr()
    """

    # The code to be tested
    result = repr(obj)

    assert re.match(r'^WBEMListener\(', result)

    exp_host_str = _format('_host={0!A}', obj.host)
    assert exp_host_str in result

    exp_http_port_str = _format('_http_port={0!A}', obj.http_port)
    assert exp_http_port_str in result

    exp_https_port_str = _format('_https_port={0!A}', obj.https_port)
    assert exp_https_port_str in result

    exp_certfile_str = _format('_certfile={0!A}', obj.certfile)
    assert exp_certfile_str in result

    exp_keyfile_str = _format('_keyfile={0!A}', obj.keyfile)
    assert exp_keyfile_str in result

    exp_logger_str = _format('_logger={0!A}', obj.logger)
    assert exp_logger_str in result


@log_entry_exit
def test_WBEMListener_start_stop():
    """
    Test starting and stopping of the listener.
    """

    host = 'localhost'
    http_port = '50000'

    listener = WBEMListener(host, http_port)
    assert listener.http_started is False
    assert listener.https_started is False

    try:
        listener.start()
        assert listener.http_started is True
        assert listener.https_started is False

    finally:
        listener.stop()
        assert listener.http_started is False
        assert listener.https_started is False


@log_entry_exit
def test_WBEMListener_port_in_use():
    """
    Test starting the listener when port is in use by another listener.
    """

    host = 'localhost'

    # Don't use this port in other tests, to be on the safe side
    # as far as port reuse is concerned.
    http_port = '50001'

    exp_exc_type = ListenerPortError

    listener1 = WBEMListener(host, http_port)
    listener2 = WBEMListener(host, http_port)

    try:
        listener1.start()
        assert listener1.http_started is True

        try:
            # The code to be tested
            listener2.start()

        except Exception as exc:  # pylint: disable=broad-except
            # e.g. on Linux
            assert isinstance(exc, exp_exc_type)
            assert listener2.http_started is False
        else:
            # e.g. on Windows
            assert listener2.http_started is True

        # Verify that in any case, listener1 is still started
        assert listener1.http_started is True

    finally:
        listener1.stop()
        listener2.stop()


@log_entry_exit
def test_WBEMListener_context_mgr():
    """
    Test starting the listener and automatic closing in a context manager.
    """

    host = 'localhost'

    # Don't use this port in other tests, to be on the safe side
    # as far as port reuse is concerned.
    http_port = '50002'

    # The code to be tested (is the context manager)
    with WBEMListener(host, http_port) as listener1:

        # Verify that CM enter returns the listener
        assert isinstance(listener1, WBEMListener)

        listener1.start()
        assert listener1.http_started is True

    # Verify that CM exit stops the listener
    assert listener1.http_started is False

    # Verify that the TCP/IP port can be used again
    listener2 = WBEMListener(host, http_port)
    try:
        listener2.start()
        assert listener2.http_started is True
    finally:
        listener2.stop()


def create_indication_data(msg_id, sequence_number, delta_time, protocol_ver):
    """
    Create a test indication from the template and input attributes.
    """

    data_template = """<?xml version="1.0" encoding="utf-8" ?>
    <CIM CIMVERSION="2.0" DTDVERSION="2.4">
      <MESSAGE ID="{msg_id}" PROTOCOLVERSION="{protocol_ver}">
        <SIMPLEEXPREQ>
          <EXPMETHODCALL NAME="ExportIndication">
            <EXPPARAMVALUE NAME="NewIndication">
              <INSTANCE CLASSNAME="CIM_AlertIndication">
                <PROPERTY NAME="Severity" TYPE="string">
                  <VALUE>high</VALUE>
                </PROPERTY>
                <PROPERTY NAME="SequenceNumber" TYPE="string">
                  <VALUE>{sequence_number}</VALUE>
                </PROPERTY>
                <PROPERTY NAME="DELTA_TIME" TYPE="string">
                  <VALUE>{delta_time}</VALUE>
                </PROPERTY>
              </INSTANCE>
            </EXPPARAMVALUE>
          </EXPMETHODCALL>
        </SIMPLEEXPREQ>
      </MESSAGE>
    </CIM>"""

    return data_template.format(
        sequence_number=sequence_number, delta_time=delta_time,
        protocol_ver=protocol_ver, msg_id=msg_id)


# Verbosity in test_WBEMListener_send_indications(). If True show summary
# information for each test
VERBOSE_SUMMARY = False

# Global variables used to communicate between the test case function and
# the process_indication_callback() function running in context of the listener
# thread. These must be global, because in Python 2, closure variables
# cannot be modified.
RCV_COUNT = 0
RCV_ERRORS = False
CALLBACK_DELAY = 0


def process_indication_callback(indication, host):
    # pylint: disable=unused-argument
    """
    This callback function gets called by the listener when an indication is
    received.

    It tests the received indication sequence number against the RCV_COUNT
    which should catch any duplicated or missing indication, since the
    counters would no longer match in such cases.

    This function is invoked in context of the listener thread. Therefore,
    it is defined as a local function in order to be able to use the
    variables from its outer function, specifically the RCV_COUNT and
    RCV_ERRORS variables.

    For the same reason, this function does not report assertion failures
    by means of raising exceptions, but by printing a message and setting
    RCV_ERRORS to an error message, so the test function can check it and
    raise assertion errors.
    """

    # Note: Global variables that are modified must be declared global
    global RCV_COUNT  # pylint: disable=global-statement
    global RCV_ERRORS  # pylint: disable=global-statement
    # global CALLBACK_DELAY  # delay in first return from callback

    LOGGER.debug(
        "Callback function called for indication number %s: %r",
        RCV_COUNT, indication.classname)

    # Delay the callback execution. This is set if the test is
    # defined to test for full queue and fail.  It delays each callback the
    # numbr of seconds in CALLBACK_DELAY
    if CALLBACK_DELAY:
        sleep(CALLBACK_DELAY)
        LOGGER.debug(
            "Callback delayed rcv_cnt %s:  delay %s sec.",
            RCV_COUNT, CALLBACK_DELAY)  # FUTURE: remove this.

    try:

        send_count = int(indication.properties['SequenceNumber'].value)
        if send_count != RCV_COUNT:
            print("Error in process_indication_callback(): Assertion error: "
                  "Unexpected SequenceNumber in received indication: "
                  f"got {send_count}, expected {RCV_COUNT}")
            sys.stdout.flush()
            RCV_ERRORS = True

        RCV_COUNT += 1

    except Exception as exc:  # pylint: disable=broad-except
        print(f"Error in process_indication_callback(): "
              f"{exc.__class__.__name__}: {exc}")
        sys.stdout.flush()
        LOGGER.debug(
            "Error in process_indication_callback() %s: %s",
            exc.__class__.__name__, exc)
        RCV_ERRORS = True


WBEMLISTENER_SEND_INDICATIONS_TESTCASES = [
    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * send_count: CIMInstanceName object to be tested.
    #   * max_queue: Max size of queue argument.
    #   * exp_success: Boolan, True if successful completion expected. If
    #     exception expected, set False
    #   * callback_delay - number that defines delay (seconds)) in first
    #     callback execution. Required for tests of queue full.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger
    (
        "Send 1 indication",
        {
            'send_count': 1,
            'max_queue': 0,
            'exp_success': True,
            'callback_delay': 0,
        },
        None, None, OK
    ),
    (
        "Send 10 indications",
        {
            'send_count': 10,
            'max_queue': 0,
            'exp_success': True,
            'callback_delay': 2,

        },
        None, None, OK
    ),
    (
        "Send 100 indications",
        {
            'send_count': 100,
            'max_queue': 0,
            'exp_success': True,
            'callback_delay': 0,

        },
        None, None, OK
    ),
    (
        "Test queue full. Send 5 indications but fail at number 2",
        {
            'send_count': 5,
            'max_queue': 2,
            'exp_success': False,
            'callback_delay': 1,

        },
        None, None, OK
    ),
    (
        "Test queue full. Send 100 indications Full = 200, no fail",
        {
            'send_count': 100,
            'max_queue': 190,
            'exp_success': True,
            'callback_delay': 1,

        },
        None, None, OK
    ),
    (
        "Test queue full. Send 5 indications full=6. No fail",
        {
            'send_count': 5,
            'max_queue': 6,
            'exp_success': False,
            'callback_delay': 1,

        },
        None, None, OK
    ),
    (
        "Test queue full. Send 100 indications, full = 90 but fail at 90",
        {
            'send_count': 100,
            'max_queue': 90,
            'exp_success': False,
            'callback_delay': 1,

        },
        None, None, OK
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    WBEMLISTENER_SEND_INDICATIONS_TESTCASES)
@simplified_test_function
@log_entry_exit
def test_WBEMListener_send_indications(
        testcase, send_count, max_queue, exp_success, callback_delay):
    """
    Test WBEMListener with an indication generator.

    This test sends the number of indications defined by the send_count
    parameter using HTTP. It confirms that they are all received by the
    listener.

    This test validates the main paths of the listener and that the listener
    can receive large numbers of indications without duplicates or dropping
    indications.

    The test generates the indications itself and does not depend on a
    WBEM Server.  Tests in tests/end2end test with indications from
    a WBEM server.

    It does not validate all of the possible xml options on indications.

    Creates the listener, starts the listener, creates the indication XML and
    adds sequence number and time to the indication instance and sends that
    instance using requests.

    The indication instance is modified for each indication count so that each
    carries its own sequence number.
    """
    # Note: Global variables that are modified must be declared global
    global RCV_COUNT  # pylint: disable=global-statement
    global RCV_ERRORS  # pylint: disable=global-statement
    global CALLBACK_DELAY  # pylint: disable=global-statement

    CALLBACK_DELAY = callback_delay

    # Fixes issue #528 where on Windows, localhost adds multisecond delay
    # probably due to hosts table or DNS misconfiguration.
    if sys.platform == 'win32':
        host = '127.0.0.1'
    else:
        host = 'localhost'
    http_port = 50000

    listener = WBEMListener(host, http_port, max_ind_queue_size=max_queue)
    listener.add_callback(process_indication_callback)

    timer = ElapsedTimer()
    stop_indication_sender = False

    try:
        listener.start()
        start_time = time()
        url = f'http://{host}:{http_port}'
        cim_protocol_version = '1.4'
        headers = {
            'Content-Type': 'application/xml; charset=utf-8',
            'CIMExport': 'MethodRequest',
            'CIMExportMethod': 'ExportIndication',
            'Accept-Encoding': 'Identity',
            'CIMProtocolVersion': cim_protocol_version,
        }
        # We include Accept-Encoding because of requests issue.
        # He supplies it if we don't.  TODO: try None

        delta_time = time() - start_time
        random_base = randint(1, 10000)

        RCV_COUNT = 0
        RCV_ERRORS = False
        LOGGER.debug("Sending %s indications", send_count)
        for i in range(send_count):

            if stop_indication_sender:
                LOGGER.debug(
                    "Will stop sending sending indications at number %s", i)
                break

            msg_id = random_base + i
            payload = create_indication_data(msg_id, i, delta_time,
                                             cim_protocol_version)

            LOGGER.debug("Sending indication number %s", i)

            try:
                response = post_bsl(url, headers=headers, data=payload)
            except requests.exceptions.RequestException as exc:
                msg = (f"Sending indication {i} "
                       f"raised {exc.__class__.__name__}: {exc}")
                LOGGER.error(msg)
                # If testing for fail with max_queue, stop sending if
                # exception from sender and do not execute AssertionError
                if max_queue and i >= max_queue:
                    LOGGER.debug("Stop sending indications at number %s: "
                                 "max_queue_size=%s", i, max_queue)
                    break

                new_exec = AssertionError(msg)
                # Disable to see original traceback
                new_exec.__cause__ = None
                raise new_exec

            LOGGER.debug(
                "Received response from sending indication number %s", i)

            if response.status_code != 200:
                msg = (f"Sending indication {i} failed with HTTP "
                       f"status {response.status_code}")
                LOGGER.error(msg)
                # Ignore error if testing for queue full error
                if not stop_indication_sender:
                    raise AssertionError(msg)

        # Indications sent.
        # Confirm that the listener thread has processed all indications. Waits
        # for indication queue to empty.
        # Cancel any callback delays so callbacks just clear queue.
        CALLBACK_DELAY = 0

        # max seconds to wait for queue to empty
        # NOTE: This does not have any short circuit if the callback processor
        # hangs up and does not finish.  It just runs out the timer.
        queue_empty_retries = max((send_count * 5), 30)

        # Wait loop to allow receive queue to empty.
        empty = False
        for i in range(queue_empty_retries):
            sleep(0.2)   # sleep 200 ms to allow callbacks to execute
            if listener.ind_queue_exists():
                qsize = listener.ind_queue_size()
                empty = listener.ind_queue_empty()
            else:
                qsize = None
                empty = None
            LOGGER.debug(
                "Waiting for empty or deinitialized queue: qsize=%s empty=%s "
                "retries=%s", qsize, empty, i)
            if empty is None or empty is True:
                LOGGER.debug("Break from wait loop.")
                break

            LOGGER.debug("Waiting, %s still in queue.",
                         listener.ind_queue_size())

        assert empty is None or empty is True, \
            f"Test failed in wait for indications loop. rcv_count=" \
            f"{RCV_COUNT} still in queue={qsize}"

        # Test for receive error, and correct receive count
        # Ignore rcvd count asserts if exp_success is False.
        if exp_success:
            assert not RCV_ERRORS, \
                f"Errors occurred in process_indication_callback(), as " \
                f" printed to stdout. testcase={testcase}."

            assert send_count == RCV_COUNT, \
                f"Mismatch between total send count {send_count} and " \
                f" receive count {RCV_COUNT}. testcase={testcase}."

    except ListenerQueueFullError as qfe:
        if exp_success:
            LOGGER.debug("ListenerQueueFullError received: %s", qfe)
            assert False, "Unexpxpected ListenerQueueFullError. Test Fail"
        else:
            LOGGER.debug("Expected and received ListenerQueueFullError")
            stop_indication_sender = True
            listener.stop()
            LOGGER.debug("Testcase fail expected result OK, listener closed.")

    finally:

        LOGGER.debug("Start listener.stop")
        listener.stop()
        LOGGER.debug("End listener.stop")

        endtime = timer.elapsed_sec()
        ind_per_sec = send_count / endtime

        msg = (f"SUMMARY: Sent {send_count} indications in {endtime:.2f} sec; "
               f"{ind_per_sec:.2f} ind/sec")

        LOGGER.info(msg)

        if VERBOSE_SUMMARY:
            print(f"\n{msg}")
            sys.stdout.flush()

        # Test that only the main thread exists.
        assert threading.active_count() == 1


@pytest.mark.parametrize(
    "method, exp_status", [
        ('OPTIONS', 405),
        ('HEAD', 405),
        ('GET', 405),
        ('PUT', 405),
        ('PATCH', 405),
        ('DELETE', 405),
        ('TRACE', 405),
        ('CONNECT', 405),
        ('M_POST', 405),
    ])
@log_entry_exit
def test_WBEMListener_incorrect_method(method, exp_status):
    """
    Verify that WBEMListener send fails when an incorrect HTTP method is used.
    """

    host = 'localhost'
    http_port = 50000
    url = f'http://{host}:{http_port}'
    headers = {
        'Content-Type': 'application/xml; charset=utf-8',
        'CIMExport': 'MethodRequest',
        'CIMExportMethod': 'ExportIndication',
        'Accept-Encoding': 'Identity',
        'CIMProtocolVersion': '1.4',
    }

    listener = WBEMListener(host, http_port)

    try:
        listener.add_callback(process_indication_callback)
        listener.start()

        # The code to be tested is running in listener thread
        response = requests.request(method, url, headers=headers, timeout=4)

        assert response.status_code == exp_status

    finally:
        listener.stop()


WBEMLISTENER_INCORRECT_HEADERS_TESTCASES = [
    (
        "Invalid Accept-Charset header",
        {
            'Content-Type': 'application/xml',
            'Accept-Charset': 'ASCII',
            'CIMExport': 'MethodRequest',
            'CIMExportMethod': 'ExportIndication',
            'Accept-Encoding': 'Identity',
            'CIMProtocolVersion': '1.4',
        },
        406,
        {
            'CIMErrorDetails': r'Invalid Accept-Charset header .*',
            'CIMError': r'header-mismatch',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "Accept-Range header not permitted",
        {
            'Content-Type': 'application/xml',
            'Accept-Range': 'foo',
            'CIMExport': 'MethodRequest',
            'CIMExportMethod': 'ExportIndication',
            'Accept-Encoding': 'Identity',
            'CIMProtocolVersion': '1.4',
        },
        406,
        {
            'CIMErrorDetails': r'Accept-Range header is not permitted .*',
            'CIMError': r'header-mismatch',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "Missing Content-Type header",
        {
            'CIMExport': 'MethodRequest',
            'CIMExportMethod': 'ExportIndication',
            'Accept-Encoding': 'Identity',
            'CIMProtocolVersion': '1.4',
        },
        406,
        {
            'CIMErrorDetails': r'Content-Type header is required',
            'CIMError': r'header-mismatch',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "Invalid Content-Type header",
        {
            'Content-Type': 'foo_application/xml',
            'CIMExport': 'MethodRequest',
            'CIMExportMethod': 'ExportIndication',
            'Accept-Encoding': 'Identity',
            'CIMProtocolVersion': '1.4',
        },
        406,
        {
            'CIMErrorDetails': r'Invalid Content-Type header .*',
            'CIMError': r'header-mismatch',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "Invalid Content-Encoding header",
        {
            'Content-Type': 'application/xml',
            'Content-Encoding': 'foo',
            'CIMExport': 'MethodRequest',
            'CIMExportMethod': 'ExportIndication',
            'Accept-Encoding': 'Identity',
            'CIMProtocolVersion': '1.4',
        },
        406,
        {
            'CIMErrorDetails': r'Invalid Content-Encoding header .*',
            'CIMError': r'header-mismatch',
            'CIMExport': r'MethodResponse',
        }
    ),
]


@pytest.mark.parametrize(
    "desc, headers, exp_status, exp_headers",
    WBEMLISTENER_INCORRECT_HEADERS_TESTCASES)
@log_entry_exit
def test_WBEMListener_incorrect_headers(desc, headers, exp_status, exp_headers):
    # pylint: disable=unused-argument
    """
    Verify that WBEMListener send fails when incorrect HTTP headers are used
    (along with the correct POST method).
    """

    host = 'localhost'
    http_port = 50000
    url = f'http://{host}:{http_port}'
    # headers = copy(headers)

    listener = WBEMListener(host, http_port)

    try:
        listener.add_callback(process_indication_callback)
        listener.start()

        # The code to be tested is running in listener thread
        response = post_bsl(url, headers=headers, data=None)

        assert response.status_code == exp_status
        for header_name in exp_headers:
            assert header_name in response.headers
            exp_header_pattern = exp_headers[header_name]
            assert re.match(exp_header_pattern, response.headers[header_name])

    finally:
        listener.stop()


WBEMLISTENER_INCORRECT_PAYLOAD1_TESTCASES = [
    (
        "Ill-formed XML in payload",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">""",
        400,
        {
            'CIMErrorDetails': r'XML parsing error .*',
            'CIMError': r'request-not-well-formed',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "Unsupported DTD version (DTDVERSION in CIM element)",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="1.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
            <SIMPLEEXPREQ>
              <EXPMETHODCALL NAME="ExportIndication">
              </EXPMETHODCALL>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        400,
        {
            'CIMErrorDetails': r'DTDVERSION is 1.4, expected 2.x.y',
            'CIMError': r'unsupported-dtd-version',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "Unsupported protocol version (PROTOCOLVERSION in MESSAGE element)",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="2.4">
            <SIMPLEEXPREQ>
              <EXPMETHODCALL NAME="ExportIndication">
              </EXPMETHODCALL>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        400,
        {
            'CIMErrorDetails': r'PROTOCOLVERSION is 2.4, expected 1.x.y',
            'CIMError': r'unsupported-protocol-version',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "CIM-XML: Missing MESSAGE element",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
        </CIM>""",
        400,
        {
            'CIMErrorDetails': r'Element .CIM. is missing required '
                               r'child element .*.MESSAGE..*',
            'CIMError': r'request-not-well-formed',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "CIM-XML: Missing SIMPLEEXPREQ element",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
          </MESSAGE>
        </CIM>""",
        400,
        {
            'CIMErrorDetails': r'Element .MESSAGE. is missing required '
                               r'child element .*.SIMPLEEXPREQ..*',
            'CIMError': r'request-not-well-formed',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "CIM-XML: Missing EXPMETHODCALL element",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
            <SIMPLEEXPREQ>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        400,
        {
            'CIMErrorDetails': r'Element .SIMPLEEXPREQ. is missing required '
                               r'child element .*.EXPMETHODCALL..*',
            'CIMError': r'request-not-well-formed',
            'CIMExport': r'MethodResponse',
        }
    ),
    (
        "CIM-XML: Invalid parameter type in ExportIndication export request",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
            <SIMPLEEXPREQ>
              <EXPMETHODCALL NAME="ExportIndication">
                <EXPPARAMVALUE NAME="NewIndication">
                  <INSTANCENAME CLASSNAME="CIM_AlertIndication">
                  </INSTANCENAME>
                </EXPPARAMVALUE>
              </EXPMETHODCALL>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        400,
        {
            'CIMErrorDetails': r'Element .EXPPARAMVALUE. has invalid child '
                               r'element.*INSTANCENAME.*',
            'CIMError': r'request-not-well-formed',
            'CIMExport': r'MethodResponse',
        }
    ),
]


@pytest.mark.parametrize(
    "desc, payload, exp_status, exp_headers",
    WBEMLISTENER_INCORRECT_PAYLOAD1_TESTCASES)
@log_entry_exit
def test_WBEMListener_incorrect_payload1(
        desc, payload, exp_status, exp_headers):
    # pylint: disable=unused-argument
    """
    Verify that WBEMListener send fails with HTTP error when incorrect HTTP
    payload is used that triggers HTTP errors.
    """

    host = 'localhost'
    http_port = 50000
    url = f'http://{host}:{http_port}'
    headers = {
        'Content-Type': 'application/xml; charset=utf-8',
        'CIMExport': 'MethodRequest',
        'CIMExportMethod': 'ExportIndication',
        'Accept-Encoding': 'Identity',
        'CIMProtocolVersion': '1.4',
    }

    listener = WBEMListener(host, http_port)

    try:
        listener.add_callback(process_indication_callback)
        listener.start()

        # The code to be tested is running in listener thread
        response = post_bsl(url, headers=headers, data=payload)

        assert response.status_code == exp_status
        for header_name in exp_headers:
            assert header_name in response.headers
            exp_header_pattern = exp_headers[header_name]
            assert re.match(exp_header_pattern, response.headers[header_name])

    finally:
        listener.stop()


WBEMLISTENER_INCORRECT_PAYLOAD2_TESTCASES = [
    (
        "Unsupported export method in export request",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
            <SIMPLEEXPREQ>
              <EXPMETHODCALL NAME="fooExportIndication">
              </EXPMETHODCALL>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        200,
        {
            'CIMExport': r'MethodResponse',
        },
        b'<?xml version="1.0" encoding="utf-8" ?>\n'
        b'<CIM CIMVERSION="2.0" DTDVERSION="2.4">'
        b'<MESSAGE ID="42" PROTOCOLVERSION="1.4">'
        b'<SIMPLEEXPRSP>'
        b'<EXPMETHODRESPONSE NAME="fooExportIndication">'
        b'<ERROR CODE="7" DESCRIPTION="Unknown export method: '
        b'\'fooExportIndication\'"/>'
        b'</EXPMETHODRESPONSE>'
        b'</SIMPLEEXPRSP>'
        b'</MESSAGE>'
        b'</CIM>'
    ),
    (
        "No parameters in ExportIndication export request",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
            <SIMPLEEXPREQ>
              <EXPMETHODCALL NAME="ExportIndication">
              </EXPMETHODCALL>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        200,
        {
            'CIMExport': r'MethodResponse',
        },
        b'<?xml version="1.0" encoding="utf-8" ?>\n'
        b'<CIM CIMVERSION="2.0" DTDVERSION="2.4">'
        b'<MESSAGE ID="42" PROTOCOLVERSION="1.4">'
        b'<SIMPLEEXPRSP>'
        b'<EXPMETHODRESPONSE NAME="ExportIndication">'
        b'<ERROR CODE="7" DESCRIPTION="Expecting one parameter '
        b'NewIndication.*"/>'
        b'</EXPMETHODRESPONSE>'
        b'</SIMPLEEXPRSP>'
        b'</MESSAGE>'
        b'</CIM>'
    ),
    (
        "Two parameters in ExportIndication export request",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
            <SIMPLEEXPREQ>
              <EXPMETHODCALL NAME="ExportIndication">
                <EXPPARAMVALUE NAME="NewIndication">
                  <INSTANCE CLASSNAME="CIM_AlertIndication">
                  </INSTANCE>
                </EXPPARAMVALUE>
                <EXPPARAMVALUE NAME="foo">
                  <INSTANCE CLASSNAME="CIM_AlertIndication">
                  </INSTANCE>
                </EXPPARAMVALUE>
              </EXPMETHODCALL>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        200,
        {
            'CIMExport': r'MethodResponse',
        },
        b'<?xml version="1.0" encoding="utf-8" ?>\n'
        b'<CIM CIMVERSION="2.0" DTDVERSION="2.4">'
        b'<MESSAGE ID="42" PROTOCOLVERSION="1.4">'
        b'<SIMPLEEXPRSP>'
        b'<EXPMETHODRESPONSE NAME="ExportIndication">'
        b'<ERROR CODE="7" DESCRIPTION="Expecting one parameter '
        b'NewIndication.*"/>'
        b'</EXPMETHODRESPONSE>'
        b'</SIMPLEEXPRSP>'
        b'</MESSAGE>'
        b'</CIM>'
    ),
    (
        "Invalid parameter name in ExportIndication export request",
        """<?xml version="1.0" encoding="utf-8" ?>
        <CIM CIMVERSION="2.0" DTDVERSION="2.4">
          <MESSAGE ID="42" PROTOCOLVERSION="1.4">
            <SIMPLEEXPREQ>
              <EXPMETHODCALL NAME="ExportIndication">
                <EXPPARAMVALUE NAME="fooNewIndication">
                  <INSTANCE CLASSNAME="CIM_AlertIndication">
                  </INSTANCE>
                </EXPPARAMVALUE>
              </EXPMETHODCALL>
            </SIMPLEEXPREQ>
          </MESSAGE>
        </CIM>""",
        200,
        {
            'CIMExport': r'MethodResponse',
        },
        b'<?xml version="1.0" encoding="utf-8" ?>\n'
        b'<CIM CIMVERSION="2.0" DTDVERSION="2.4">'
        b'<MESSAGE ID="42" PROTOCOLVERSION="1.4">'
        b'<SIMPLEEXPRSP>'
        b'<EXPMETHODRESPONSE NAME="ExportIndication">'
        b'<ERROR CODE="7" DESCRIPTION="Expecting one parameter '
        b'NewIndication.*"/>'
        b'</EXPMETHODRESPONSE>'
        b'</SIMPLEEXPRSP>'
        b'</MESSAGE>'
        b'</CIM>'
    ),
]


@pytest.mark.parametrize(
    "desc, payload, exp_status, exp_headers, exp_payload",
    WBEMLISTENER_INCORRECT_PAYLOAD2_TESTCASES)
@log_entry_exit
def test_WBEMListener_incorrect_payload2(
        desc, payload, exp_status, exp_headers, exp_payload):
    # pylint: disable=unused-argument
    """
    Verify that WBEMListener send fails with export response indicating error
    when incorrect HTTP payload is used that triggers that.
    """

    host = 'localhost'
    http_port = 50000
    url = f'http://{host}:{http_port}'
    headers = {
        'Content-Type': 'application/xml; charset=utf-8',
        'CIMExport': 'MethodRequest',
        'CIMExportMethod': 'ExportIndication',
        'Accept-Encoding': 'Identity',
        'CIMProtocolVersion': '1.4',
    }

    listener = WBEMListener(host, http_port)

    try:
        listener.add_callback(process_indication_callback)
        listener.start()

        # The code to be tested is running in listener thread
        response = post_bsl(url, headers=headers, data=payload)

        assert response.status_code == exp_status
        for header_name in exp_headers:
            assert header_name in response.headers
            exp_header_pattern = exp_headers[header_name]
            assert re.match(exp_header_pattern, response.headers[header_name])
        act_payload = response.content
        re.match(exp_payload, act_payload, re.MULTILINE)

    finally:
        listener.stop()


# Attribute value in testcase definition indicating not to compare the value
NOCHECK_VALUE = -999

THREAD_INIT_TESTCASES = [
    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * thread_class: Thread class to be tested.
    # * kwargs: Keyword arguments for the Thread class init.
    # * exp_attrs: Expected attributes.
    #   Key: Attribute name. Value: tuple(type, value). value can be
    #   NOCHECK_VALUE to indicate not to check the value.

    (
        "ExceptionHandlingThread",
        ExceptionHandlingThread,
        dict(
            target=None,
            name='foo',
        ),
        dict(
            name=(str, 'foo'),
            exception=(NoneType, None),
        )
    ),
    (
        "StoppableThread",
        StoppableThread,
        dict(
            target=None,
            name='foo',
        ),
        dict(
            name=(str, 'foo'),
            stop_event=(threading.Event, NOCHECK_VALUE),
        )
    ),
    (
        "ServerThread",
        ServerThread,
        dict(
            target=None,
            name='foo',
        ),
        dict(
            name=(str, 'foo'),
            exception=(NoneType, None),
        )
    ),
    (
        "CallbackThread",
        CallbackThread,
        dict(
            target=None,
            name='foo',
        ),
        dict(
            name=(str, 'foo'),
            exception=(NoneType, None),
            stop_event=(threading.Event, NOCHECK_VALUE),
        )
    ),
]


@pytest.mark.parametrize(
    "desc, thread_class, kwargs, exp_attrs",
    THREAD_INIT_TESTCASES)
def test_thread_init(
        desc, thread_class, kwargs, exp_attrs):
    # pylint: disable=unused-argument
    """
    Verify that the thread class shows the expected attributes after init.
    """

    # The code to be tested
    thread = thread_class(**kwargs)

    for name, exp_type_value in exp_attrs.items():
        exp_type, exp_value = exp_type_value
        assert hasattr(thread, name)
        value = getattr(thread, name)
        # pylint: disable=unidiomatic-typecheck)
        assert type(value) == exp_type  # noqa: E721
        if exp_value != NOCHECK_VALUE:
            assert value == exp_value


class StoppableThreadFuncHolder:
    # pylint: disable=too-few-public-methods
    """
    Test support class providing the thread function, that solves the problem
    for the thread function to access the thread to check for the stop
    condition.
    """

    def __init__(self):
        self.thread = None

    def thread_func(self, interval):
        """
        Thread function that regularly checks the thread whether to stop.
        """
        LOGGER.debug("thread_func: Called with interval=%.2f", interval)
        assert self.thread is not None
        while True:
            if self.thread.stopped():
                break
            LOGGER.debug("thread_func: Sleeping for %.2f s", interval)
            sleep(interval)
        LOGGER.debug("thread_func: Returning")


@pytest.mark.parametrize(
    "thread_class",
    [
        StoppableThread,
        CallbackThread,  # derived from StoppableThread
    ]
)
def test_thread_stoppable(thread_class):
    """
    Verify that a StoppableThread is stoppable.
    """
    assert issubclass(thread_class, StoppableThread)

    LOGGER.debug("test_thread_stoppable: Test function called with "
                 "thread_class=%r", thread_class)

    func_holder = StoppableThreadFuncHolder()
    thread_interval = 0.1

    # Create a thread as a daemon thread. This will cause it to be cleaned up
    # when the Python process running pytest terminates.
    thread = thread_class(
        target=func_holder.thread_func,
        kwargs=dict(interval=thread_interval),
        name='foo',
        daemon=True)
    func_holder.thread = thread

    LOGGER.debug("test_thread_stoppable: Starting thread")
    thread.start()

    sleep_time = 3 * thread_interval
    LOGGER.debug("test_thread_stoppable: Sleeping for %.2f s", sleep_time)
    sleep(sleep_time)

    LOGGER.debug("test_thread_stoppable: Stopping thread")
    # The code to be tested
    thread.stop()

    # The code to be tested
    stopped_result = thread.stopped()

    # Verify that stop() has been attempted
    assert stopped_result is True

    # Wait for thread to end
    join_timeout = 2 * sleep_time
    LOGGER.debug("test_thread_stoppable: Waiting for thread to end within "
                 "%.2f s", join_timeout)
    thread.join(join_timeout)

    # Verify that the thread no longer exists
    assert not thread.is_alive()
    LOGGER.debug("test_thread_stoppable: As expected, the thread ended")

    LOGGER.debug("test_thread_stoppable: Test function finished")


class ExcHdlThreadFuncHolder:
    # pylint: disable=too-few-public-methods
    """
    Test support class providing the thread function, that solves the problem
    for the thread function to access the thread to check for the stop
    condition.
    """

    def __init__(self):
        self.thread = None

    def thread_func(self, exc):
        """
        Thread function that raises an exception.
        """
        LOGGER.debug("thread_func: Called with exc=%r", exc)
        assert self.thread is not None

        if exc:
            LOGGER.debug("thread_func: Raising exception %r", exc)
            raise exc

        LOGGER.debug("thread_func: Returning without raising exception")


@pytest.mark.parametrize(
    "exc",
    [
        ValueError("foobar"),
        None,
    ]
)
@pytest.mark.parametrize(
    "thread_class",
    [
        ExceptionHandlingThread,
        ServerThread,  # derived from ExceptionHandlingThread
        CallbackThread,  # derived from ExceptionHandlingThread
    ]
)
def test_thread_exchdl(thread_class, exc):
    """
    Verify that a ExceptionHandlingThread handles exceptions.
    """
    assert issubclass(thread_class, ExceptionHandlingThread)

    LOGGER.debug("test_thread_exchdl: Test function called with "
                 "thread_class=%r, exc=%r", thread_class, exc)

    func_holder = ExcHdlThreadFuncHolder()

    # Create a thread as a daemon thread. This will cause it to be cleaned up
    # when the Python process running pytest terminates.
    thread = thread_class(
        target=func_holder.thread_func,
        kwargs=dict(exc=exc),
        name='foo',
        daemon=True)
    func_holder.thread = thread

    LOGGER.debug("test_thread_exchdl: Starting thread")
    thread.start()

    # Wait for thread to end
    join_timeout = 5
    LOGGER.debug("test_thread_exchdl: Waiting for thread to end within "
                 "%.2f s", join_timeout)

    if exc:

        # Verify that join() raises the specified exception
        with pytest.raises(type(exc)) as exc_info:
            thread.join(join_timeout)
        act_exc = exc_info.value
        assert act_exc == exc
        LOGGER.debug("test_thread_exchdl: As expected, thread.join() raised "
                     "exception %r", exc)

    else:

        # Verify that join() does not raise any exception
        thread.join(join_timeout)
        LOGGER.debug("test_thread_exchdl: As expected, thread.join() did not "
                     "raise an exception")

    # Verify that the thread no longer exists
    assert not thread.is_alive()
    LOGGER.debug("test_thread_exchdl: As expected, the thread ended")

    # Verify that the thread has stored the exception
    assert thread.exception is exc

    LOGGER.debug("test_thread_exchdl: Test function finished")



#!/usr/bin/env python
"""
Test cases for _tupletree module and Unicode/XML check functions.
"""

import xml
import re

import pytest

from ..utils.pytest_extensions import log_entry_exit

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import _tupletree, ParseError  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


# The following was originally the pywbem parser.  Now it is used simply as a
# check on the SAX parser.  Therefore, this code was moved from _tupletree.py to
# here for the tests. Note that this code is referenced from other code in the
# attic so we might not want to eliminate it completely.

def dom_to_tupletree(node):
    """
    Convert a DOM object to a pyRXP-style tuple tree.

    Each element is a 4-tuple of (NAME, ATTRS, CONTENTS, None).

    Very nice for processing complex nested trees.
    """

    if node.nodeType == node.DOCUMENT_NODE:
        # boring; pop down one level
        return dom_to_tupletree(node.firstChild)
    assert node.nodeType == node.ELEMENT_NODE

    name = node.nodeName
    attrs = {}
    contents = []

    for child in node.childNodes:
        if child.nodeType == child.ELEMENT_NODE:
            contents.append(dom_to_tupletree(child))
        elif child.nodeType == child.TEXT_NODE:
            assert isinstance(child.nodeValue, str), \
                f"text node is not a string: {child}"
            contents.append(child.nodeValue)
        elif child.nodeType == child.CDATA_SECTION_NODE:
            contents.append(child.nodeValue)
        else:
            raise RuntimeError(f"can't handle {child}")

    for i in range(node.attributes.length):
        attr_node = node.attributes.item(i)
        attrs[attr_node.nodeName] = attr_node.nodeValue

    # TODO: Cannot handle comments, cdata, processing instructions, etc.

    # it's so easy in retrospect!
    return (name, attrs, contents)


def xml_to_tupletree(xml_string):
    """
    Parse XML straight into tupletree.
    Uses the minidom to parse xml_string int a dom object.
    This is part of the old obsolete dom code
    """
    dom_xml = xml.dom.minidom.parseString(xml_string)
    return dom_to_tupletree(dom_xml)


class Test_xml_to_tupletree_sax:
    # pylint: disable=too-few-public-methods
    """
    Exhaustive tests for _tupletree.xml_to_tupletree_sax(), with inline input
    XML and expected tupletrees (for success) or exceptions (for failures).

    A comparison with the result of the previously implemented DOM based
    pywbem parser ensures that the SAX parsing is compatible.
    """

    testcases = [
        # Each testcase has these items:
        # * desc: Short testcase description.
        # * xml_string: XML string to be parsed, as unicode string.
        # * exp_tupletree: Expected tupletree parse result,
        #   or None if expecting failure.
        # * exp_exc_type: Expected exception type,
        #   or None if expecting success.
        # * exp_exc_msg_pattern: Expected pattern in exception message,
        #   or None if expecting success.
        # * condition: Condition for testcase to run.

        # General good cases
        (
            "One-line XML string with one empty element",
            '<?xml version="1.0" encoding="utf-8" ?>'
            '<TEXT></TEXT>',
            ('TEXT',
             {},
             []),
            None, None,
            True
        ),
        (
            "One-line XML string with one empty element, short form",
            '<?xml version="1.0" encoding="utf-8" ?>'
            '<TEXT/>',
            ('TEXT',
             {},
             []),
            None, None,
            True
        ),
        (
            "One-line XML string with one element with ASCII text",
            '<?xml version="1.0" encoding="utf-8" ?>'
            '<TEXT>foo</TEXT>',
            ('TEXT',
             {},
             ['foo']),
            None, None,
            True
        ),
        (
            "One-line XML string with one element with non-ASCII UCS-2 text",
            '<?xml version="1.0" encoding="utf-8" ?>'
            '<TEXT>\u00e8</TEXT>',
            ('TEXT',
             {},
             ['\u00e8']),
            None, None,
            True
        ),
        (
            "One-line XML string with one element with non-UCS-2 text",
            '<?xml version="1.0" encoding="utf-8" ?>'
            '<TEXT>\U00010142</TEXT>',  # GREEK ACROPHONIC ATTIC ONE DRACHMA
            ('TEXT',
             {},
             ['\U00010142']),
            None, None,
            True
        ),
        (
            "One-line XML string with one element with one attribute",
            '<?xml version="1.0" encoding="utf-8" ?>'
            '<TEXT ATTR="foo">bar</TEXT>',
            ('TEXT',
             {'ATTR': 'foo'},
             ['bar']),
            None, None,
            True
        ),
        (
            "One-line XML string with one element with two attributes",
            '<?xml version="1.0" encoding="utf-8" ?>'
            '<TEXT ATTR1="foo1" ATTR2="foo2">bar</TEXT>',
            ('TEXT',
             {'ATTR1': 'foo1', 'ATTR2': 'foo2'},
             ['bar']),
            None, None,
            True
        ),
        (
            "One-line XML string with two elements and two attributes",
            '<?xml version="1.0" encoding="utf-8" ?>'
            '<TEXT1 ATTR1="foo1"><TEXT2 ATTR2="foo2">bar</TEXT2></TEXT1>',
            ('TEXT1',
             {'ATTR1': 'foo1'},
             [('TEXT2',
               {'ATTR2': 'foo2'},
               ['bar'])]),
            None, None,
            True
        ),

        # Multi-line good cases
        (
            "Multi-line XML string, text without newline",
            '<?xml version="1.0" encoding="utf-8" ?>\n'
            '<TEXT1 ATTR1="foo1">\n'
            '<TEXT2 ATTR2="foo2">bar</TEXT2>\n'
            '</TEXT1>',
            ('TEXT1',
             {'ATTR1': 'foo1'},
             ['\n',
              ('TEXT2',
               {'ATTR2': 'foo2'},
               ['bar']),
              '\n']),
            None, None,
            True
        ),
        (
            "Multi-line XML string, text with newlines",
            '<?xml version="1.0" encoding="utf-8" ?>\n'
            '<TEXT1 ATTR1="foo1">\n'
            '<TEXT2 ATTR2="foo2">\n'
            'bar\n'
            '</TEXT2>\n'
            '</TEXT1>',
            ('TEXT1',
             {'ATTR1': 'foo1'},
             ['\n',
              ('TEXT2',
               {'ATTR2': 'foo2'},
               ['\nbar\n']),
              '\n']),
            None, None,
            True
        ),

        # General failure cases
        (
            "Empty XML string",
            '',
            None,
            ParseError, "XML parsing error.*no element found",
            True
        ),
        (
            "Just the XML prolog",
            '<?xml version="1.0" encoding="utf-8" ?>',
            None,
            ParseError, "XML parsing error.*no element found",
            True
        ),
        (
            "Start element without end element",
            '<?xml version="1.0" encoding="utf-8" ?>'
            '<ELEM>',
            None,
            ParseError, "XML parsing error.*no element found",
            True
        ),
        (
            "End element misses trailing angle bracket",
            '<?xml version="1.0" encoding="utf-8" ?>'
            '<ELEM>abc</ELEM',
            None,
            ParseError, "XML parsing error.*unclosed token",
            True
        ),
        (
            "End element misses leading angle bracket",
            '<?xml version="1.0" encoding="utf-8" ?>'
            '<ELEM>abc/ELEM>',
            None,
            ParseError, "XML parsing error.*no element found",
            True
        ),
        (
            "Start element misses trailing angle bracket",
            '<?xml version="1.0" encoding="utf-8" ?>'
            '<ELEM</ELEM>',
            None,
            ParseError, "XML parsing error.*not well-formed",
            True
        ),
        (
            "Start element misses leading angle bracket",
            '<?xml version="1.0" encoding="utf-8" ?>'
            'ELEM></ELEM>',
            None,
            ParseError, "XML parsing error.*syntax error",
            True
        ),

        # Failure cases with line number checking
        (
            "Failing multi-line XML string with failure in line 3",
            '<?xml version="1.0" encoding="utf-8" ?>\n'
            '<ELEM1>\n'
            '<ELEM2</ELEM2>\n'
            '<ELEM1>\n',
            None,
            ParseError,
            "XML parsing error.*not well-formed.*"
            "Line 3 column 6 of XML string.*"
            "<ELEM2</ELEM2>",
            True
        ),

        # Failure cases with non-ASCII chars
        (
            "Failing XML string with non-ASCII attribute value",
            '<?xml version="1.0" encoding="utf-8" ?>'
            '<ELEM1 ATTR1="\u00e8"></ELEM2>',
            None,
            ParseError,
            "XML parsing error.*mismatched tag.*"
            "Line 1 column 58 of XML string.*"
            r'<ELEM1 ATTR1="\\xc3\\xa8"></ELEM2>',
            True
        ),

        # CIM-XML cases
        (
            "CIM-XML response with MESSAGE end element missing "
            "(former Associators_error.xml file)",
            '<?xml version="1.0" encoding="utf-8" ?>\n'
            '<CIM CIMVERSION="2.0" DTDVERSION="2.0">\n'
            '  <MESSAGE ID="1001" PROTOCOLVERSION="1.0">\n'
            '    <SIMPLERSP>\n'
            '      <IMETHODRESPONSE NAME="Associators">\n'
            '        <IRETURNVALUE>\n'
            '        </IRETURNVALUE>\n'
            '      </IMETHODRESPONSE>\n'
            '    </SIMPLERSP>\n'
            '</CIM>\n',
            None,
            ParseError, "XML parsing error.*mismatched tag",
            True
        ),
        (
            "CIM-XML response from Associators operation "
            "(former Associators_Empty.xml)",
            '<?xml version="1.0" encoding="utf-8" ?>\n'
            '<CIM CIMVERSION="2.0" DTDVERSION="2.0">\n'
            '  <MESSAGE ID="1001" PROTOCOLVERSION="1.0">\n'
            '    <SIMPLERSP>\n'
            '      <IMETHODRESPONSE NAME="Associators">\n'
            '        <IRETURNVALUE>\n'
            '        </IRETURNVALUE>\n'
            '      </IMETHODRESPONSE>\n'
            '    </SIMPLERSP>\n'
            '  </MESSAGE>\n'
            '</CIM>\n',
            ('CIM',
             {'CIMVERSION': '2.0', 'DTDVERSION': '2.0'},
             ['\n  ',
              ('MESSAGE',
               {'ID': '1001', 'PROTOCOLVERSION': '1.0'},
               ['\n    ',
                ('SIMPLERSP',
                 {},
                 ['\n      ',
                  ('IMETHODRESPONSE',
                   {'NAME': 'Associators'},
                   ['\n        ',
                    ('IRETURNVALUE',
                     {},
                     ['\n        ']),
                    '\n      ']),
                  '\n    ']),
                '\n  ']),
              '\n']),
            None, None,
            True
        ),
        (
            "CIM-XML response from Associators operation "
            "(former Associators_StorageVolume_small.xml)",
            '<?xml version="1.0" encoding="utf-8" ?>\n'
            '<CIM CIMVERSION="2.0" DTDVERSION="2.0"><MESSAGE ID="1001" '
            'PROTOCOLVERSION="1.0">\n'
            '<SIMPLERSP><IMETHODRESPONSE NAME="Associators"><IRETURNVALUE>\n'
            '    \n'
            '<VALUE.OBJECTWITHPATH>\n'
            '<INSTANCEPATH>\n'
            '<NAMESPACEPATH>\n'
            '<HOST>10.10.10.10</HOST>\n'
            '<LOCALNAMESPACEPATH>\n'
            '<NAMESPACE NAME="root"/>\n'
            '<NAMESPACE NAME="emc"/>\n'
            '</LOCALNAMESPACEPATH>\n'
            '</NAMESPACEPATH>\n'
            '<INSTANCENAME CLASSNAME="Symm_StorageVolume">\n'
            '<KEYBINDING NAME="CreationClassName"><KEYVALUE '
            'VALUETYPE="string" TYPE="string">Symm_StorageVolume</KEYVALUE>'
            '</KEYBINDING>\n'
            '<KEYBINDING NAME="DeviceID"><KEYVALUE VALUETYPE="string" '
            'TYPE="string">00000</KEYVALUE></KEYBINDING>\n'
            '<KEYBINDING NAME="SystemCreationClassName"><KEYVALUE '
            'VALUETYPE="string" TYPE="string">Symm_StorageSystem</KEYVALUE>'
            '</KEYBINDING>\n'
            '<KEYBINDING NAME="SystemName"><KEYVALUE VALUETYPE="string" '
            'TYPE="string">SYMMETRIX-+-000194900000</KEYVALUE></KEYBINDING>\n'
            '</INSTANCENAME>\n'
            '</INSTANCEPATH>\n'
            '<INSTANCE CLASSNAME="Symm_StorageVolume">\n'
            '<PROPERTY NAME="Usage" TYPE="uint16"><VALUE>3</VALUE>\n'
            '</PROPERTY>\n'
            '</INSTANCE>\n'
            '</VALUE.OBJECTWITHPATH>\n'
            '\n'
            '</IRETURNVALUE></IMETHODRESPONSE></SIMPLERSP></MESSAGE></CIM>\n',
            ('CIM',
             {'CIMVERSION': '2.0', 'DTDVERSION': '2.0'},
             [('MESSAGE',
               {'ID': '1001', 'PROTOCOLVERSION': '1.0'},
               ['\n',
                ('SIMPLERSP',
                 {},
                 [('IMETHODRESPONSE',
                   {'NAME': 'Associators'},
                   [('IRETURNVALUE',
                     {},
                     ['\n    \n',
                      ('VALUE.OBJECTWITHPATH',
                       {},
                       ['\n',
                        ('INSTANCEPATH',
                         {},
                         ['\n',
                          ('NAMESPACEPATH',
                           {},
                           ['\n',
                            ('HOST',
                             {},
                             ['10.10.10.10']),
                            '\n',
                            ('LOCALNAMESPACEPATH',
                             {},
                             ['\n',
                              ('NAMESPACE',
                               {'NAME': 'root'},
                               []),
                              '\n',
                              ('NAMESPACE',
                               {'NAME': 'emc'},
                               []),
                              '\n']),
                            '\n']),
                          '\n',
                          ('INSTANCENAME',
                           {'CLASSNAME': 'Symm_StorageVolume'},
                           ['\n',
                            ('KEYBINDING',
                             {'NAME': 'CreationClassName'},
                             [('KEYVALUE',
                               {'TYPE': 'string',
                                'VALUETYPE': 'string'},
                               ['Symm_StorageVolume'])]),
                            '\n',
                            ('KEYBINDING',
                             {'NAME': 'DeviceID'},
                             [('KEYVALUE',
                               {'TYPE': 'string',
                                'VALUETYPE': 'string'},
                               ['00000'])]),
                            '\n',
                            ('KEYBINDING',
                             {'NAME': 'SystemCreationClassName'},
                             [('KEYVALUE',
                               {'TYPE': 'string',
                                'VALUETYPE': 'string'},
                               ['Symm_StorageSystem'])]),
                            '\n',
                            ('KEYBINDING',
                             {'NAME': 'SystemName'},
                             [('KEYVALUE',
                               {'TYPE': 'string',
                                'VALUETYPE': 'string'},
                               ['SYMMETRIX-+-000194900000'])]),
                            '\n']),
                          '\n']),
                        '\n',
                        ('INSTANCE',
                         {'CLASSNAME': 'Symm_StorageVolume'},
                         ['\n',
                          ('PROPERTY',
                           {'NAME': 'Usage',
                            'TYPE': 'uint16'},
                           [('VALUE',
                             {},
                             ['3']),
                            '\n']),
                          '\n']),
                        '\n']),
                      '\n\n'])])])])]),
            None, None,
            True
        ),
    ]

    @pytest.mark.parametrize(
        "desc, xml_string, exp_tupletree, exp_exc_type, exp_exc_msg_pattern, "
        "condition",
        testcases
    )
    @pytest.mark.parametrize(
        "encoding",
        ['unicode', 'bytes']
    )
    @log_entry_exit
    def test_xml_to_tupletree_sax(
            self, encoding, desc, xml_string, exp_tupletree, exp_exc_type,
            exp_exc_msg_pattern, condition):
        # pylint: disable=no-self-use,unused-argument
        """
        Test xml_to_tupletree_sax() against expected results and against
        DOM parser results.
        """

        if not condition:
            pytest.skip("Condition for test case not met")

        assert isinstance(xml_string, str)
        if encoding == 'bytes':
            xml_string = xml_string.encode("utf-8")
        else:
            assert encoding == 'unicode'

        if exp_exc_type is None:

            act_tupletree = _tupletree.xml_to_tupletree_sax(
                xml_string, 'Test XML')

            # Compare against expected result
            assert act_tupletree == exp_tupletree

            if encoding == 'bytes':
                # Compare against result from previously used DOM parser
                dom_tupletree = xml_to_tupletree(xml_string)
                assert act_tupletree == dom_tupletree

        else:
            with pytest.raises(exp_exc_type) as exec_info:

                _tupletree.xml_to_tupletree_sax(xml_string, 'Test XML')

            if exp_exc_msg_pattern:
                exc = exec_info.value
                exc_msg = str(exc)
                one_line_exc_msg = exc_msg.replace('\n', '\\n')
                assert re.search(exp_exc_msg_pattern, one_line_exc_msg), \
                    "Unexpected exception message:\n" + exc_msg


class Test_get_failing_line:
    # pylint: disable=too-few-public-methods
    """Tests for _tupletree.get_failing_line()"""

    testcases = [
        # Each testcase has these items:
        # * desc: Short testcase description.
        # * xml_string: Input XML string for the function.
        # * exc_msg: Input exception message for the function.
        # * exp_result: Expected result tuple, or None if expecting failure.
        # * exp_exc_type: Expected exception type, or None if expecting success.
        # * condition: Condition for testcase to run.

        # General cases
        (
            "Short one line XML string, parseable exception message",
            b'<V>ab</V>',
            ":1:2: some message",
            (1, 2, 3, "'<V>ab</V>'"),
            None, True
        ),
        (
            "Short 2 line XML string, parseable exception message",
            b'<V>ab</V>\n'
            b'<X>ab</X>\n',
            ":2:2: some message",
            (2, 2, 3, "'<X>ab</X>'"),
            None, True
        ),
        (
            "Short 2 line XML string, non-parseable exception message",
            b'<V>ab</V>\n'
            b'<X>ab</X>\n',
            "(2:2) some message",
            (None, None, None, "'<V>ab</V>\\n<X>ab</X>\\n'"),
            None, True
        ),
        (
            "Short 2 line XML string, parseable exception message for line 3",
            b'<V>ab</V>\n'
            b'<X>ab</X>\n',
            ":3:2: some message",
            (None, None, None, "'<V>ab</V>\\n<X>ab</X>\\n'"),
            None, True
        ),
        (
            "Empty XML, parseable exception message for line 1",
            b'',
            ":1:1: empty string",
            (1, 1, 2, "''"),
            None, True
        ),
        (
            "Empty line with newline XML, parseable exc. message for line 1",
            b'\n',
            ":1:1: empty string",
            (1, 1, 2, "''"),
            None, True
        ),

        # Exceeding chars after colno, with 0 chars before colno
        (
            "XML with 0/x/500 chars",
            b'b' * 0 + b'x' + b'a' * 500,
            ":1:1: empty string",
            (1, 1, 2, "'" + 'b' * 0 + 'x' + 'a' * 500 + "'"),
            None, True
        ),
        (
            "XML with 0/x/501 chars",
            b'b' * 0 + b'x' + b'a' * 501,
            ":1:1: empty string",
            (1, 1, 2, "'" + 'b' * 0 + 'x' + 'a' * 500 + "'" + '...'),
            None, True
        ),

        # Exceeding chars after colno, with 5 chars before colno
        (
            "XML with 5/x/500 chars",
            b'b' * 5 + b'x' + b'a' * 500,
            ":1:6: empty string",
            (1, 6, 7, "'" + 'b' * 5 + 'x' + 'a' * 500 + "'"),
            None, True
        ),
        (
            "XML with 5/x/501 chars",
            b'b' * 5 + b'x' + b'a' * 501,
            ":1:6: empty string",
            (1, 6, 7, "'" + 'b' * 5 + 'x' + 'a' * 500 + "'" + '...'),
            None, True
        ),

        # Exceeding chars before colno, with 0 chars after colno
        (
            "XML with 500/x/0 chars",
            b'b' * 500 + b'x' + b'a' * 0,
            ":1:501: empty string",
            (1, 501, 502, "'" + 'b' * 500 + 'x' + 'a' * 0 + "'"),
            None, True
        ),
        (
            "XML with 501/x/0 chars",
            b'b' * 501 + b'x' + b'a' * 0,
            ":1:502: empty string",
            (1, 502, 505, '...' + "'" + 'b' * 500 + 'x' + 'a' * 0 + "'"),
            None, True
        ),

        # Exceeding chars before colno, with 5 chars after colno
        (
            "XML with 500/x/5 chars",
            b'b' * 500 + b'x' + b'a' * 5,
            ":1:501: empty string",
            (1, 501, 502, "'" + 'b' * 500 + 'x' + 'a' * 5 + "'"),
            None, True
        ),
        (
            "XML with 501/x/5 chars",
            b'b' * 501 + b'x' + b'a' * 5,
            ":1:502: empty string",
            (1, 502, 505, '...' + "'" + 'b' * 500 + 'x' + 'a' * 5 + "'"),
            None, True
        ),

        # Exceeding chars before and after colno
        (
            "XML with 500/x/500 chars",
            b'b' * 500 + b'x' + b'a' * 500,
            ":1:501: empty string",
            (1, 501, 502, "'" + 'b' * 500 + 'x' + 'a' * 500 + "'"),
            None, True
        ),
        (
            "XML with 501/x/501 chars",
            b'b' * 501 + b'x' + b'a' * 501,
            ":1:502: empty string",
            (1, 502, 505, '...' + "'" + 'b' * 500 + 'x' + 'a' * 500 + "'" +
             '...'),
            None, True
        ),

    ]

    @pytest.mark.parametrize(
        "desc, xml_string, exc_msg, exp_result, exp_exc_type, condition",
        testcases
    )
    @log_entry_exit
    def test_get_failing_line(
            self, desc, xml_string, exc_msg, exp_result, exp_exc_type,
            condition):
        # pylint: disable=no-self-use,unused-argument
        """Tests for _tupletree.get_failing_line()"""

        if not condition:
            pytest.skip("Condition for test case not met")

        if exp_exc_type is None:

            act_result = _tupletree.get_failing_line(xml_string, exc_msg)

            assert isinstance(act_result, tuple)
            assert act_result == exp_result

        else:
            with pytest.raises(exp_exc_type):

                _tupletree.get_failing_line(xml_string, exc_msg)


class Test_check_invalid_utf8_sequences:
    # pylint: disable=too-few-public-methods
    """
    Tests for _tupletree.check_invalid_utf8_sequences()
    """

    testcases = [
        # Each testcase has these items:
        # * desc: Short testcase description.
        # * utf8_string: Input string for the function.
        # * exp_exc_type: Expected exception type, or None if expecting success.
        # * condition: Condition for testcase to run.

        # General cases
        (
            "Incorrect type of input string",
            '<V>ab</V>',
            TypeError, True
        ),

        # Good cases
        (
            "Good case without special chars",
            b'<V>a</V>',
            None, True
        ),
        (
            "Good case with U+0009 (TAB), U+000A (NL), U+000D (CR) chars",
            b'<V>a\x09b\x0Ac\x0Dd</V>',
            None, True
        ),
        (
            "Good case with U+0350 char (in UTF-8)",
            b'<V>a\xCD\x90b</V>',
            None, True
        ),
        (
            "Good case with U+2013 char (in UTF-8)",
            b'<V>a\xE2\x80\x93b</V>',
            None, True
        ),
        (
            "Good case with U+10122 char (in UTF-8)",
            b'<V>a\xF0\x90\x84\xA2b</V>',
            None, True
        ),

        # Correctly encoded but ill-formed UTF-8
        (
            "Ill-formed UTF-8 using surrogate U+D800,U+DD22 "
            "(in otherwise correctly encoded UTF-8)",
            b'<V>a\xED\xA0\x80\xED\xB4\xA2b</V>',
            ParseError, True
        ),
        (
            "Ill-formed UTF-8 using surrogates U+D800,U+DD22 and U+D800,U+DD23 "
            "(in otherwise correctly encoded UTF-8)",
            b'<V>a\xED\xA0\x80\xED\xB4\xA2b\xED\xA0\x80\xED\xB4\xA3</V>',
            ParseError, True
        ),

        # Incorrectly encoded UTF-8
        (
            "Incorrectly encoded UTF-8 with 1-byte sequence",
            b'<V>a\x80b</V>',
            ParseError, True
        ),
        (
            "Incorrectly encoded UTF-8 with 2-byte sequence "
            "with missing second byte",
            b'<V>a\xC0',
            ParseError, True
        ),
        (
            "Incorrectly encoded UTF-8 with 2-byte sequence "
            "with incorrect 2nd byte",
            b'<V>a\xC0b</V>',
            ParseError, True
        ),
        (
            "Incorrectly encoded UTF-8 with 4-byte sequence "
            "with incorrect 3rd byte",
            b'<V>a\xF1\x80abc</V>',
            ParseError, True
        ),
        (
            "Incorrectly encoded UTF-8 with 4-byte sequence "
            "with incorrect 3rd byte that is an incorrect new start",
            b'<V>a\xF1\x80\xFFbc</V>',
            ParseError, True
        ),
        (
            "Incorrectly encoded UTF-8 with 4-byte sequence "
            "with incorrect 3rd byte that is a correct new start",
            b'<V>a\xF1\x80\xC2\x81c</V>',
            ParseError, True
        ),
    ]

    @pytest.mark.parametrize(
        "desc, utf8_string, exp_exc_type, condition",
        testcases
    )
    @log_entry_exit
    def test_check_invalid_utf8_sequences(
            self, desc, utf8_string, exp_exc_type, condition):
        # pylint: disable=no-self-use
        """
        Tests for _tupletree.check_invalid_utf8_sequences()
        """

        if not condition:
            pytest.skip("Condition for test case not met")

        if exp_exc_type is None:

            unicode_string = _tupletree.check_invalid_utf8_sequences(
                utf8_string, desc)

            assert isinstance(unicode_string, str), desc
            utf8_string_u = utf8_string.decode('utf-8')
            assert unicode_string == utf8_string_u, desc

        else:
            with pytest.raises(exp_exc_type):

                _tupletree.check_invalid_utf8_sequences(utf8_string, desc)


class Test_check_invalid_xml_chars:
    # pylint: disable=too-few-public-methods
    """
    Tests for _tupletree.check_invalid_xml_chars()
    """

    testcases = [
        # Each testcase has these items:
        # * desc: Short testcase description.
        # * xml_string: Input string for the function.
        # * exp_exc_type: Expected exception type, or None if expecting success.
        # * condition: Condition for testcase to run.

        # General cases
        (
            "Incorrect type of input string",
            b'<V>ab</V>',
            TypeError, True
        ),

        # Good cases
        (
            "Good case without special chars",
            '<V>a</V>',
            None, True
        ),
        (
            "Good case with U+0009 (TAB), U+000A (NL), U+000D (CR) chars",
            '<V>a\u0009b\u000Ac\u000Dd</V>',
            None, True
        ),
        (
            "Good case with U+0350 char",
            '<V>a\u0350b</V>',
            None, True
        ),
        (
            "Good case with U+2013 char",
            '<V>a\u2013b</V>',
            None, True
        ),
        (
            "Good case with U+10122 char",
            '<V>a\u010122b</V>',
            None, True
        ),

        # Invalid XML characters
        (
            "Invalid XML char U+0008 (BEL)",
            '<V>a\u0008b</V>',
            ParseError, True
        ),
        (
            "Invalid XML char U+0000",
            '<V>a\u0000b</V>',
            ParseError, True
        ),
        (
            "Invalid XML chars U+0001 ... U+0008",
            '<V>a\u0001b\u0002c\u0003d\u0004e\u0005f\u0006g\u0007h\u0008i</V>',
            ParseError, True
        ),
        (
            "Invalid XML chars U+000B ... U+000C",
            '<V>a\u000Bb\u000Cc</V>',
            ParseError, True
        ),
        (
            "Invalid XML chars U+000E ... U+0015",
            '<V>a\u000Eb\u000Fc\u0010d\u0011e\u0012f\u0013g\u0014h\u0015i</V>',
            ParseError, True
        ),
        (
            "Invalid XML chars U+0016 ... U+001D",
            '<V>a\u0016b\u0017c\u0018d\u0019e\u001Af\u001Bg\u001Ch\u001Di</V>',
            ParseError, True
        ),
        (
            "Invalid XML chars U+001E ... U+001F",
            '<V>a\u001Eb\u001Fc</V>',
            ParseError, True
        ),
    ]

    @pytest.mark.parametrize(
        "desc, xml_string, exp_exc_type, condition",
        testcases
    )
    @log_entry_exit
    def test_check_invalid_xml_chars(
            self, desc, xml_string, exp_exc_type, condition):
        # pylint: disable=no-self-use
        """
        Tests for _tupletree.check_invalid_xml_chars()
        """

        if not condition:
            pytest.skip("Condition for test case not met")

        if exp_exc_type is None:

            _tupletree.check_invalid_xml_chars(xml_string, desc)

        else:
            with pytest.raises(exp_exc_type):

                _tupletree.check_invalid_xml_chars(xml_string, desc)
